__all__ = ['Replacement']

from typing import Dict, List, Sequence
import functools

from .location import FileLocationRange


# FIXME: use attrs
class Replacement(object):
    """
    Describes the replacement of a contiguous body of text in a single source
    code file with a provided text.
    """
    @staticmethod
    def from_dict(d: Dict[str, str]) -> 'Replacement':
        location = FileLocationRange.from_string(d['location'])
        return Replacement(location, d['text'])

    @staticmethod
    def resolve(replacements: Sequence['Replacement']
                ) -> List['Replacement']:
        """
        Resolves all conflicts in a sequence of replacements.
        """
        # group by file
        file_to_reps = {}  # type: Dict[str, List[Replacement]]
        for rep in replacements:
            if rep.filename not in file_to_reps:
                file_to_reps[rep.filename] = []
            file_to_reps[rep.filename].append(rep)

        # resolve redundant replacements
        for fn in file_to_reps:
            reps = file_to_reps[fn]

            def cmp(x, y) -> int:
                return -1 if x < y else 0 if x == y else 0

            def compare(x, y) -> int:
                start_x, stop_x = x.location.start, x.location.stop
                start_y, stop_y = y.location.start, y.location.stop
                if start_x != start_y:
                    return cmp(start_x, start_y)
                # start_x == start_y
                return -cmp(stop_x, stop_y)

            reps.sort(key=functools.cmp_to_key(compare))

            filtered = [reps[0]]  # type: List[Replacement]
            i, j = 0, 1
            while j < len(reps):
                x, y = reps[i], reps[j]
                if x.location.stop > y.location.start:
                    j += 1
                else:
                    i += 1
                    j += 1
                    filtered.append(y)
            file_to_reps[fn] = filtered

        # collapse into a flat sequence of transformations
        resolved = []  # type: List[Replacement]
        for reps in file_to_reps.values():
            resolved += reps
        return resolved

    def __init__(self,
                 location: FileLocationRange,
                 text: str
                 ) -> None:
        self.__location = location
        self.__text = text

    def __repr__(self) -> str:
        return "Replacement({}, {})".format(repr(self.location),
                                            self.text)

    @property
    def filename(self) -> str:
        """
        The name of the file in which the replacement should be made.
        """
        return self.__location.filename

    @property
    def location(self) -> FileLocationRange:
        """
        The contiguous range of text that should be replaced.
        """
        return self.__location

    @property
    def text(self) -> str:
        """
        The source text that should be used as a replacement.
        """
        return self.__text

    def to_dict(self) -> Dict[str, str]:
        return {'location': str(self.location),
                'text': self.text}



from typing import Any, List, Dict, Iterable, Iterator

import attr

from bugzoo.core.fileline import FileLine, FileLineSet

__all__ = ['Location', 'LocationRange', 'FileLocationRange', 'FileLine',
           'FileLocationRangeSet', 'FileLineSet']


@attr.s(frozen=True, repr=False)
class Location(object):
    """
    Represents a character location within an arbitrary file.
    """
    @staticmethod
    def from_string(s: str) -> 'Location':
        line, _, col = s.partition(':')
        return Location(int(line), int(col))

    line = attr.ib(type=int)
    col = attr.ib(type=int)

    @property
    def column(self) -> int:
        return self.col

    def __le__(self, other: 'Location') -> bool:
        if self.line == other.line:
            return self.col < other.col
        return self.line < other.line

    def __str__(self) -> str:
        return "{}:{}".format(self.line, self.column)

    def __repr__(self) -> str:
        return "Location({})".format(str(self))


@attr.s(frozen=True, repr=False)
class FileLocation(object):
    """
    Represents a character location within a named file.
    """
    filename = attr.ib(type=str)
    location = attr.ib(type=Location)

    @property
    def col(self) -> int:
        return self.location.col

    @property
    def line(self) -> int:
        return self.location.line

    @staticmethod
    def from_string(s: str) -> 'FileLocation':
        filename, _, loc_s = s.rpartition('@')
        location = Location.from_string(loc_s)
        return FileLocation(filename, location)

    def __str__(self) -> str:
        return "{}@{}".format(self.filename, self.location)

    def __repr__(self) -> str:
        return "FileLocation({})".format(str(self))


@attr.s(frozen=True, repr=False)
class LocationRange(object):
    """
    Captures a continuous range of source code locations.
    """
    @staticmethod
    def from_string(s: str) -> 'LocationRange':
        start_s, _, stop_s = s.partition("::")
        start = Location.from_string(start_s)
        stop = Location.from_string(stop_s)
        return LocationRange(start, stop)

    start = attr.ib(type=Location)
    stop = attr.ib(type=Location)

    def __str__(self) -> str:
        return "{}::{}".format(self.start, self.stop)

    def __repr__(self) -> str:
        return "LocationRange({})".format(str(self))

    def __contains__(self, loc: Location) -> bool:
        """
        Determines whether a given location is contained within this range.
        """
        left = loc.line > self.start.line \
            or (loc.line == self.start.line and loc.col >= self.start.col)
        right = loc.line < self.stop.line \
            or (loc.line == self.stop.line and loc.col < self.stop.col)
        return left and right


@attr.s(frozen=True, repr=False)
class FileLocationRange(object):
    """
    Represents a contiguous sequence of characters in a particular file.
    """
    @staticmethod
    def from_string(s: str) -> 'FileLocationRange':
        filename, _, s_range = s.rpartition('@')
        location_range = LocationRange.from_string(s_range)
        return FileLocationRange(filename, location_range)

    filename = attr.ib(type=str)
    location_range = attr.ib(type=LocationRange)

    @property
    def start(self) -> Location:
        return self.location_range.start

    @property
    def stop(self) -> Location:
        return self.location_range.stop

    def __str__(self) -> str:
        return "{}@{}".format(self.filename, self.location_range)

    def __repr__(self) -> str:
        return "FileLocationRange({})".format(str(self))

    def __contains__(self, floc: FileLocation) -> bool:
        """
        Determines whether a given location is contained within this range.
        """
        in_file = floc.filename == self.filename
        in_range = floc.location in self.location_range
        return in_file and in_range


class FileLocationRangeSet(object):
    def __init__(self, ranges: Iterable[FileLocationRange]) -> None:
        """
        Represents a set of file location ranges.
        """
        self.__fn_to_ranges = {}  # type: Dict[str, List[FileLocationRange]]
        for r in ranges:
            if r.filename not in self.__fn_to_ranges:
                self.__fn_to_ranges[r.filename] = []
            self.__fn_to_ranges[r.filename].append(r)

    def __iter__(self) -> Iterator[FileLocationRange]:
        for f_ranges in self.__fn_to_ranges.values():
            yield from f_ranges

    def __repr__(self) -> str:
        return "FileLocationRangeSet({})".format(self.__fn_to_ranges)

    def __contains__(self, location: FileLocation) -> bool:
        """
        Determines whether a given file location is contained in one of the
        ranges within this set.
        """
        ranges = self.__fn_to_ranges.get(location.filename, [])
        return any(location in r for r in ranges)
